import calendar
import holidays
import json
import os
import pandas as pd
import re
import requests
from datetime import datetime
from enum import Enum
from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from webdriver_manager.chrome import ChromeDriverManager

date_format = "%d/%m/%Y"
days_of_week = {
    "Monday": 0,
    "Tuesday": 1,
    "Wednesday": 2,
    "Thursday": 3,
    "Friday": 4,
    "Saturday": 5,
    "Sunday": 6,
}


class Region(Enum):
    ENG = 1
    NIR = 2
    SCT = 3
    WLS = 4


def check_postcode(postcode: str):
    """
    Checks a postcode exists and validates UK formatting against a RegEx string
        :param postcode: Postcode to parse
    """
    postcode_api_url = "https://api.postcodes.io/postcodes/"
    postcode_api_response = requests.get(f"{postcode_api_url}{postcode}")

    if postcode_api_response.status_code != 200:
        val_error = json.loads(postcode_api_response.text)
        raise ValueError(
            f"Exception: {val_error['error']} Status: {val_error['status']}"
        )
    return True


def check_paon(paon: str):
    """
    Checks that PAON data exists
        :param paon: PAON data to check, usually house number
    """
    try:
        if paon is None:
            raise ValueError("Invalid house number")
        return True
    except Exception as ex:
        print(f"Exception encountered: {ex}")
        print("Please check the provided house number.")
        exit(1)


def check_uprn(uprn: str):
    """
    Checks that the UPRN exists
        :param uprn: UPRN to check
    """
    try:
        if uprn is None or uprn == "":
            raise ValueError("Invalid UPRN")
        return True
    except Exception as ex:
        print(f"Exception encountered: {ex}")
        print("Please check the provided UPRN.")


def check_usrn(usrn: str):
    """
    Checks that the USRN exists
        :param uprn: USRN to check
    """
    try:
        if usrn is None or usrn == "":
            raise ValueError("Invalid USRN")
        return True
    except Exception as ex:
        print(f"Exception encountered: {ex}")
        print("Please check the provided USRN.")


def get_date_with_ordinal(date_number: int) -> str:
    """
    Return ordinal text on day of date
        :rtype: str
        :param date_number: Date number as an integer (e.g. 4)
        :return: Return date with ordinal suffix (e.g. 4th)
    """
    return str(date_number) + (
        "th"
        if 4 <= date_number % 100 <= 20
        else {1: "st", 2: "nd", 3: "rd"}.get(date_number % 10, "th")
    )


def remove_ordinal_indicator_from_date_string(date_string: str) -> str:
    """
    Remove the ordinal indicator from a written date as a string.
    E.g. June 12th 2022 -> June 12 2022
    :rtype: str
    """
    ord_day_pattern = re.compile(r"(?<=\d)(st|nd|rd|th)")
    return re.compile(ord_day_pattern).sub("", date_string)


def parse_header(raw_header: str) -> dict:
    """
    Parses a header string and returns one that can be useful
            :rtype: dict
            :param raw_header: header as a string, with values to separate as pipe (|)
            :return: header in a dictionary format that can be used in requests
    """
    header = dict()
    for line in raw_header.split("|"):
        if line.startswith(":"):
            a, b = line[1:].split(":", 1)
            a = f":{a}"
        else:
            a, b = line.split(":", 1)

        header[a.strip()] = b.strip()

    return header


def is_holiday(date_to_check: datetime, region: Region = Region.ENG) -> bool:
    """
    Checks if a given date is a public holiday
        :param date_to_check: Date to check if holiday
        :param region: The UK nation to check. Defaults to ENG.
        :return: Bool - true if a holiday, false if not
    """
    uk_holidays = holidays.country_holidays("GB", subdiv=region.name)

    if date_to_check in uk_holidays:
        return True
    else:
        return False


def get_weekday_dates_in_period(start: datetime, day_of_week: int, amount=8) -> list:
    """
    Returns a list of dates of a given weekday from a start date for the given amount of weeks
        :param start: Start date
        :param day_of_week: Day of week number. Recommended to use calendar.DAY (Monday=0, Sunday=6)
        :param amount: Number of weeks to get dates. Defaults to 8 weeks.
        :return: List of dates where the specified weekday is in the period
    """
    return (
        pd.date_range(
            start=start, freq=f"W-{calendar.day_abbr[day_of_week]}", periods=amount
        )
        .strftime(date_format)
        .tolist()
    )


def get_dates_every_x_days(start: datetime, step: int, amount: int = 8) -> list:
    """
    Returns a list of dates for `X` days from start date. For example, calling `get_stepped_dates_in_period(s, 21, 4)` would
    return `4` dates every `21` days from the start date `s`
        :param start: Date to start from
        :param step: X amount of days
        :param amount: Number of dates to find
        :return: List of dates every X days from start date
        :rtype: list
    """
    return (
        pd.date_range(start=start, freq=f"{step}D", periods=amount)
        .strftime(date_format)
        .tolist()
    )


def get_next_occurrence_from_day_month(date: datetime) -> datetime:
    day = datetime.now().date().strftime("%d")
    month = datetime.now().date().strftime("%m")
    if date.strftime("%m") < month or (date.strftime("%m") == month and date.strftime("%d") < day):
        date = pd.to_datetime(date) + pd.DateOffset(years=1)
    return date


def remove_alpha_characters(input_string: str) -> str:
    return "".join(c for c in input_string if c.isdigit() or c == " ")


def update_input_json(council: str, url: str, **kwargs):
    """
    Create/update council's entry in the input.json
        :param council: Council
        :param kwargs: Run arguments
    """
    postcode = kwargs.get("postcode", None)
    paon = kwargs.get("paon", None)
    uprn = kwargs.get("uprn", None)
    usrn = kwargs.get("usrn", None)
    web_driver = kwargs.get("web_driver", None)
    skip_get_url = kwargs.get("skip_get_url", None)
    cwd = os.getcwd()
    input_file_path = os.path.join(cwd, "uk_bin_collection", "tests", "input.json")
    if os.path.exists(input_file_path):
        with open(input_file_path, 'r') as f:
            data = json.load(f)
            if council not in data:
                data[council] = {"wiki_name": council}
            if "url" != "":
                data[council]["url"] = url
            if postcode is not None:
                data[council]["postcode"] = postcode
            if paon is not None:
                data[council]["paon"] = paon
            if uprn is not None:
                data[council]["uprn"] = uprn
            if usrn is not None:
                data[council]["usrn"] = usrn
            if web_driver is not None:
                data[council]["web_driver"] = web_driver
            if skip_get_url is not None:
                data[council]["skip_get_url"] = skip_get_url
        with open(input_file_path, 'w') as f:
            f.write(json.dumps(data, sort_keys=True, indent=4))
    else:
        print("Exception encountered: Unable to update input.json file for the council.")
        print(
            "Please check you're running developer mode from either the UKBinCollectionData "
            "or uk_bin_collection/uk_bin_collection/ directories."
        )


def validate_dates(bin_dates: dict) -> dict:
    raise NotImplementedError()
    # If a date is in December and the next is in January, increase the year


def create_webdriver(web_driver) -> webdriver.Chrome:
    """
    Create and return a headless Selenium webdriver
    :rtype: webdriver.Chrome
    """
    # Set up Selenium to run 'headless'
    options = webdriver.ChromeOptions()
    options.add_argument("--headless")
    options.add_argument("--no-sandbox")
    options.add_argument("--disable-gpu")
    options.add_argument("--disable-dev-shm-usage")
    options.add_experimental_option("excludeSwitches", ["enable-logging"])
    # Return a remote Selenium webdriver
    if web_driver is not None:
        return webdriver.Remote(command_executor=web_driver, options=options)
    # Return a local Selenium webdriver
    return webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), options=options)



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        data = {"bins": []}

        soup = BeautifulSoup(page.text, features="html.parser")
        collections = soup.find_all("h3", {"class": "waste-service-name"})
        for c in collections:
            rows = c.find_next_sibling("div", {"class": "govuk-grid-row"}).find_all(
                "div", {"class": "govuk-summary-list__row"}
            )
            for row in rows:
                if row.find("dt").get_text().strip().lower() == "next collection":
                    collection_date = (
                        remove_ordinal_indicator_from_date_string(row.find("dd").get_text()).strip()
                    )
                    dict_data = {
                        "type": c.get_text().strip().capitalize(),
                        "collectionDate": get_next_occurrence_from_day_month(datetime.strptime(
                            collection_date + " " + datetime.now().strftime("%Y"),
                            "%A, %d %B %Y",
                        )).strftime(date_format),
                    }
                    data["bins"].append(dict_data)

        data["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), date_format)
        )

        return data



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass

from datetime import timedelta
from dateutil.relativedelta import relativedelta


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Make a BS4 object
        soup = BeautifulSoup(page.text, features="html.parser")
        soup.prettify()

        data = {"bins": []}
        curr_date = datetime.today()

        for bins in soup.find_all("ul", {"class": "refuse"}):
            binCollection = bins.find_all("li")

            if binCollection:
                for bin in binCollection:
                    collection_date = bin.find("strong", {"class": "date"}).contents[0].strip()
                    if collection_date.lower() == "today":
                        collection_date = datetime.now()
                    elif collection_date.lower() == "tomorrow":
                        collection_date = datetime.now() + timedelta(days=1)
                    else:
                        collection_date = datetime.strptime(remove_ordinal_indicator_from_date_string(collection_date)
                                                            .strip(),"%a %d %b").replace(year=curr_date.year)
                        if curr_date.month == 12 and collection_date.month == 1:
                            bin_date = bin_date + relativedelta(years=1)
                    dict_data = {
                        "type": bin.find("a").contents[0],
                        "collectionDate": collection_date.strftime(date_format)
                    }

                    data["bins"].append(dict_data)

        return data



from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import Select, WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from dateutil.parser import parse

from uk_bin_collection.uk_bin_collection.common import create_webdriver, date_format
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


class CouncilClass(AbstractGetBinDataClass):
    def wait_for_element(self, driver, element_type, element: str, timeout: int = 5):
        element_present = EC.presence_of_element_located((element_type, element))
        self.wait_for_element_conditions(driver, element_present, timeout=timeout)

    def wait_for_element_conditions(self, driver, conditions, timeout: int = 5):
        try:
            WebDriverWait(driver, timeout).until(conditions)
        except TimeoutException:
            print("Timed out waiting for page to load")
            raise

    def parse_data(self, page: str, **kwargs) -> dict:
        web_driver = kwargs.get("web_driver")
        page = "https://westsuffolk-self.achieveservice.com/service/WSS_EX_Inf_Bin_Collection_Postcode_Lookup"

        # Assign user info
        user_postcode = kwargs.get("postcode")
        user_paon = kwargs.get("paon")

        # Create Selenium webdriver
        driver = create_webdriver(web_driver)
        driver.get(page)

        # Click the cookie button
        css_selector = "#close-cookie-message"
        self.wait_for_element(driver, By.CSS_SELECTOR, css_selector)
        cookie_button = driver.find_element(By.CSS_SELECTOR, css_selector)
        cookie_button.click()

        # switch to the form iframe
        iframe = driver.find_element(By.CSS_SELECTOR, "#fillform-frame-1")
        driver.switch_to.frame(iframe)

        # Wait for form to load
        xpath_selector = '//*[@id="postcode_search"]'
        self.wait_for_element(driver, By.XPATH, xpath_selector, timeout=6)

        # Send postcode
        postcode_input_box = driver.find_element(By.XPATH, xpath_selector)
        postcode_input_box.send_keys(user_postcode)
        postcode_input_box.send_keys(Keys.ENTER)

        search_for_address_button = driver.find_element(
            By.CSS_SELECTOR, "#addresssearch"
        )
        search_for_address_button.click()

        # Find address
        self.wait_for_element(
            driver, By.CSS_SELECTOR, "#selectaddr > option:nth-child(2)"
        )
        select_address_dropdown = Select(
            driver.find_element(By.CSS_SELECTOR, "#selectaddr")
        )
        if user_paon is not None:
            for option in select_address_dropdown.options:
                if user_paon in option.text:
                    select_address_dropdown.select_by_visible_text(option.text)
                    break
        else:
            # If we've not been supplied an address, pick the second entry
            select_address_dropdown.select_by_index(1)

        # Click the get schedule button (Once it's available)
        self.wait_for_element_conditions(
            driver,
            EC.element_to_be_clickable((By.CSS_SELECTOR, "#searchforcollections")),
        )
        check_schedules_button = driver.find_element(
            By.CSS_SELECTOR, "#searchforcollections"
        )
        check_schedules_button.click()

        # Grab the bin output data
        # Span is of the format:
        #   Your Next bin collections are:
        #   Black bin - Wednesday 29th November
        #    Blue bin - Wednesday 6th December
        #    Brown bin - Wednesday 6th December
        output_span_css_selector = "//span[@data-name='statictext6']"
        output_span_ec = EC.all_of(
            EC.presence_of_element_located((By.XPATH, output_span_css_selector)),
            EC.text_to_be_present_in_element(
                (By.XPATH, output_span_css_selector), "bin collections"
            ),
        )
        self.wait_for_element_conditions(driver, output_span_ec)
        data_span = driver.find_element(By.XPATH, output_span_css_selector)

        data = {"bins": []}
        for item in data_span.text.splitlines():
            if "bin" in item and " - " in item:
                bin_data = item.split(" - ")
                bin_name = bin_data[0]
                bin_date = bin_data[1]
                parsed_bin_date = parse(bin_date, fuzzy_with_tokens=True)[0]

                dict_data = {
                    "type": bin_name,
                    "collectionDate": parsed_bin_date.strftime(date_format),
                }

                data["bins"].append(dict_data)

        # Quit Selenium webdriver to release session
        driver.quit()

        return data



import json
import logging
import os
from jsonschema import validate, ValidationError

logging.basicConfig(level=logging.INFO, format='%(levelname)s - %(message)s')

def load_inputs_file(file_name):
    cwd = os.getcwd()
    with open(os.path.join(cwd, "uk_bin_collection", "tests", file_name)) as f:
        data = json.load(f)
        logging.info(f"{file_name} Input file loaded")
    return data


def load_schema_file(file_name):
    cwd = os.getcwd()
    with open(os.path.join(cwd, "uk_bin_collection", "tests", file_name)) as f:
        data = json.load(f)
        logging.info(f"{file_name} Schema file loaded")
    return data


def validate_json(json_str):
    try:
        json.loads(json_str)
    except ValueError as err:
        logging.info(f"The following error occured {err}")
        return False
    return True


def validate_json_schema(json_str, schema):
    json_data = json.loads(json_str)
    try:
        validate(instance=json_data, schema=schema)
    except ValidationError as err:
        logging.info(f"The following error occured {err}")
        logging.info(f"Data: {json_str}")
        logging.info(f"Schema: {schema}")
        return False
    return True



from bs4 import BeautifulSoup
from datetime import datetime
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber


class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        page = "https://nwleics-self.achieveservice.com/en/AchieveForms/?form_uri=sandbox-publish://AF-Process-bde93819-fa47-4bba-b094-bef375dbef0c/AF-Stage-b4ac5d55-7fb7-4c40-809f-4d1856399bed/definition.json&redirectlink=/en&cancelRedirectLink=/en&noLoginPrompt=1"

        data = {"bins": []}

        user_uprn = kwargs.get("uprn")
        user_postcode = kwargs.get("postcode")
        web_driver = kwargs.get("web_driver")
        check_uprn(user_uprn)
        check_postcode(user_postcode)
        # Create Selenium webdriver
        driver = create_webdriver(web_driver)
        driver.get(page)

        # If you bang in the house number (or property name) and postcode in the box it should find your property

        iframe_presense = WebDriverWait(driver, 30).until(
            EC.presence_of_element_located((By.ID, "fillform-frame-1"))
        )

        driver.switch_to.frame(iframe_presense)
        wait = WebDriverWait(driver, 60)
        inputElement_postcodesearch = wait.until(
            EC.element_to_be_clickable((By.NAME, "postcode_search"))
        )

        inputElement_postcodesearch.send_keys(user_postcode)
        inputElement_postcodesearch.send_keys(" ")

        # Wait for the 'Select your property' dropdown to appear and select the first result
        dropdown = wait.until(EC.element_to_be_clickable((By.NAME, "ChooseAddress")))

        dropdown_options = wait.until(
            EC.presence_of_element_located((By.CLASS_NAME, "lookup-option"))
        )

        # Create a 'Select' for it, then select the first address in the list
        # (Index 0 is "Make a selection from the list")
        dropdownSelect = Select(dropdown)
        dropdownSelect.select_by_value(str(user_uprn))

        # Wait for the 'View more' link to appear, then click it to get the full set of dates
        next_btn = wait.until(EC.element_to_be_clickable((By.CLASS_NAME, "nextText")))

        next_btn.click()

        dropdown_options = wait.until(
            EC.presence_of_element_located((By.ID, "WasteCollections"))
        )

        soup = BeautifulSoup(driver.page_source, features="html.parser")

        wasteTable = soup.find("tbody", id="WasteCollections").find_all("tr")

        for row in wasteTable:
            rowdata = row.find_all("td")
            if len(rowdata) == 3:
                # Strip the day (i.e. Monday) out of the collection date string for parsing
                dateString = " ".join(rowdata[2].text.strip().split(" ")[1:])

                data["bins"].append(
                    {
                        "type": rowdata[1].text.strip(),
                        "collectionDate": datetime.strptime(
                            dateString, "%d %B %Y"
                        ).strftime(date_format),
                    }
                )

        return data



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        data = {"bins": []}
        bin_types = {
            "240L RUBBISH BIN": "Black bin",
            "240L GARDEN BIN": "Green bin",
            "180L RECYCLING BIN": "Blue lid bin"
        }
        collections = []

        user_postcode = kwargs.get("postcode")
        user_uprn = kwargs.get("uprn")

        check_uprn(user_uprn)
        check_postcode(user_postcode)

        headers = {
            'authority': 'www.rugby.gov.uk',
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
            'accept-language': 'en-GB,en;q=0.9',
            'cache-control': 'no-cache',
            'content-type': 'application/x-www-form-urlencoded',
            # 'cookie': 'JSESSIONID=7E90CAB54B649C3DCC7F6B5DA0897C63; COOKIE_SUPPORT=true; GUEST_LANGUAGE_ID=en_GB; AWSELB=D941E98916B5759862ED6C39DA9FB3FD9880491851D200C98112ABEC3223D52B19A2A2C6B37A89D3650D44FA5728FCAFEDE7BB2592D948FFF9C7B18D76C41AF02C308B0F3A2DE17F1585E9959BCE68CC83BC3AC753; CookieControl={"necessaryCookies":[],"optionalCookies":{},"statement":{},"consentDate":1701710876715,"consentExpiry":90,"interactedWith":true,"user":"8FED8810-3C3E-4D50-A9DC-42655030B3B1"}',
            'origin': 'https://www.rugby.gov.uk',
            'pragma': 'no-cache',
            'referer': 'https://www.rugby.gov.uk/check-your-next-bin-day',
            'sec-ch-ua': '"Chromium";v="118", "Opera GX";v="104", "Not=A?Brand";v="99"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.5993.118 Safari/537.36',
        }
        params = {
            'p_p_id': 'com_placecube_digitalplace_local_waste_portlet_CollectionDayFinderPortlet',
            'p_p_lifecycle': '0',
            'p_p_state': 'normal',
            'p_p_mode': 'view',
            '_com_placecube_digitalplace_local_waste_portlet_CollectionDayFinderPortlet_mvcRenderCommandName': '/collection_day_finder/get_days',
        }
        data = {
            '_com_placecube_digitalplace_local_waste_portlet_CollectionDayFinderPortlet_formDate': f'{datetime.now().timestamp().__floor__()}',
            '_com_placecube_digitalplace_local_waste_portlet_CollectionDayFinderPortlet_postcode': f'{user_postcode}',
            '_com_placecube_digitalplace_local_waste_portlet_CollectionDayFinderPortlet_uprn': f'{user_uprn}',
        }

        response = requests.post(
            'https://www.rugby.gov.uk/check-your-next-bin-day',
            params=params,
            headers=headers,
            data=data,
        )

        soup = BeautifulSoup(response.text, features="html.parser")
        soup.prettify()

        table_rows = soup.find("table", {"class": "table"}).find("tbody").find_all("tr")

        for row in table_rows:
            row_text = row.text.strip().split('\n')
            bin_type = bin_types.get(row_text[0])
            collections.append((bin_type, datetime.strptime(row_text[1], "%A %d %b %Y")))
            collections.append((bin_type, datetime.strptime(row_text[3], "%A %d %b %Y")))


        ordered_data = sorted(collections, key=lambda x: x[1])
        data = {"bins": []}
        for item in ordered_data:
            dict_data = {"type": item[0], "collectionDate": item[1].strftime(date_format)}
            data["bins"].append(dict_data)

        return data



from time import sleep
from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait, Select

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the base
    class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        user_uprn = kwargs.get("uprn")
        user_postcode = kwargs.get("postcode")
        web_driver = kwargs.get("web_driver")
        check_uprn(user_uprn)
        check_postcode(user_postcode)

        # Create Selenium webdriver
        driver = create_webdriver(web_driver)
        driver.get("https://my.eastsuffolk.gov.uk/service/Bin_collection_dates_finder")

        # Wait for iframe to load and switch to it
        WebDriverWait(driver, 30).until(EC.frame_to_be_available_and_switch_to_it((By.ID, 'fillform-frame-1')))

        # Wait for postcode entry box
        postcode = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "alt_postcode_search"))
        )
        # Enter postcode
        postcode.send_keys(user_postcode.replace(" ", ""))

        # Wait for address selection dropdown to appear
        address = Select(
            WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.ID, "alt_choose_address"))
            )
        )

        # Wait for spinner to disappear (signifies options are loaded for select)
        WebDriverWait(driver, 10).until(
            EC.invisibility_of_element_located((By.CLASS_NAME, "spinner-outer"))
        )

        # Sometimes the options aren't fully there despite the spinner being gone, wait another 2 seconds.
        sleep(2)

        # Select address by UPRN
        address.select_by_value(user_uprn)

        # Wait for spinner to disappear (signifies data is loaded)
        WebDriverWait(driver, 10).until(
            EC.invisibility_of_element_located((By.CLASS_NAME, "spinner-outer"))
        )

        sleep(2)

        # Find data table
        data_table = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located(
                (
                    By.XPATH,
                    '//div[@data-field-name="collection_details"]/div[contains(@class, "fieldContent")]/div[contains(@class, "repeatable-table-wrapper")]',
                )
            )
        )

        
        # Make a BS4 object
        soup = BeautifulSoup(data_table.get_attribute("innerHTML"), features="html.parser")

        # Quit Selenium webdriver to release session
        driver.quit()

        data = {"bins": []}

        rows = soup.find("table").find("tbody").find_all("tr")
        for row in rows:
            cols = row.find_all("td")
            bin_type = cols[2].find_all("span")[1].text.title()
            collection_date = cols[3].find_all("span")[1].text
            collection_date = datetime.strptime(collection_date, "%d/%m/%Y").strftime(
                date_format
            )
            dict_data = {"type": bin_type, "collectionDate": collection_date}
            data["bins"].append(dict_data)

        data["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), date_format)
        )

        return data



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        data = {"bins": []}

        # Get UPRN and postcode from commandline
        user_uprn = kwargs.get("uprn")
        user_postcode = kwargs.get("postcode")
        check_postcode(user_postcode)
        check_uprn(user_uprn)

        # Pass in form data and make the POST request
        headers = {
            'authority': 'waste.barnsley.gov.uk',
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
            'accept-language': 'en-GB,en;q=0.9',
            'cache-control': 'no-cache',
            'content-type': 'application/x-www-form-urlencoded',
            'origin': 'https://waste.barnsley.gov.uk',
            'pragma': 'no-cache',
            'referer': 'https://waste.barnsley.gov.uk/ViewCollection/SelectAddress',
            'sec-ch-ua': '"Chromium";v="118", "Opera GX";v="104", "Not=A?Brand";v="99"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.5993.118 Safari/537.36',
        }
        form_data = {
            'personInfo.person1.HouseNumberOrName': '',
            'personInfo.person1.Postcode': f'{user_postcode}',
            'personInfo.person1.UPRN': f'{user_uprn}',
            'person1_SelectAddress': 'Select address',
        }
        response = requests.post('https://waste.barnsley.gov.uk/ViewCollection/SelectAddress', headers=headers,
                                 data=form_data)
        soup = BeautifulSoup(response.text, features="html.parser")
        soup.prettify()

        if response.status_code != 200:
            raise ConnectionRefusedError("Error getting results from website! Please open an issue on GitHub!")

        # Parse the response, getting the top box first and then tabled collections after
        results = soup.find("div", {"class": "panel"}).find_all("fieldset")[0:2]
        heading = results[0].find_all("p")[1:3]
        bin_text = heading[1].text.strip() + " bin"
        bin_date = datetime.strptime(heading[0].text, "%A, %B %d, %Y")
        dict_data = {
            "type": bin_text,
            "collectionDate": bin_date.strftime(date_format),
        }
        data["bins"].append(dict_data)

        results_table = [row for row in results[1].find_all("tbody")[0] if row != '\n']
        for row in results_table:
            text_list = [item.text.strip() for item in row.contents if item != '\n']
            bin_text = text_list[1] + " bin"
            bin_date = datetime.strptime(text_list[0], "%A, %B %d, %Y")
            dict_data = {
                "type": bin_text,
                "collectionDate": bin_date.strftime(date_format),
            }
            data["bins"].append(dict_data)

        return data



import re
import requests
from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        data = {"bins": []}
        headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64)"}

        # Make a BS4 object
        soup = BeautifulSoup(page.text, features="html.parser")
        soup.prettify()

        # Get collection calendar
        calendar_urls = soup.find_all("a", string=re.compile(r"view or download the collection calendar"))
        if len(calendar_urls) > 0:
            requests.packages.urllib3.disable_warnings()
            response = requests.get(calendar_urls[0].get("href"), headers=headers)

            # Make a BS4 object
            soup = BeautifulSoup(response.text, features="html.parser")
            soup.prettify()

            # Loop the months
            for month in soup.find_all("div", {"class": "usercontent"}):
                year = ""
                if month.find("h2") and 'calendar' not in month.find("h2").get_text(strip=True):
                    year = datetime.strptime(month.find("h2").get_text(strip=True), "%B %Y").strftime("%Y")
                elif month.find("h3"):
                    year = datetime.strptime(month.find("h3").get_text(strip=True), "%B %Y").strftime("%Y")
                if year != "":
                    for row in month.find_all("li"):
                        results = re.search(
                            "([A-Za-z]+ \\d\\d? [A-Za-z]+): (.+)", row.get_text(strip=True)
                        )
                        if results:
                            dict_data = {
                                "type": results.groups()[1].capitalize(),
                                "collectionDate": datetime.strptime(
                                    results.groups()[0] + " " + year, "%A %d %B %Y"
                                ).strftime(date_format),
                            }
                            data["bins"].append(dict_data)

            # Sort collections
            data["bins"].sort(
                key=lambda x: datetime.strptime(x.get("collectionDate"), "%d/%m/%Y")
            )

        return data



from bs4 import BeautifulSoup
from datetime import datetime
import requests
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass

COLLECTION_KINDS = {
    "waste": "rteelem_ctl03_pnlCollections_Refuse",
    "recycling": "rteelem_ctl03_pnlCollections_Recycling",
    "glass": "rteelem_ctl03_pnlCollections_Glass",
    # Garden waste data is only returned if the property is subscribed to the Garden Waste service
    "garden": "rteelem_ctl03_pnlCollections_GardenWaste"
}


class CouncilClass(AbstractGetBinDataClass):
    def parse_data(self, page: str, **kwargs) -> dict:
        requests.packages.urllib3.disable_warnings()

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)

        cookies = {
            'cookie_control_popup': 'A',
            'WhenAreMyBinsCollected': f'{user_uprn}',
        }

        headers = {
            'Accept': '*/*',
            'Accept-Language': 'en-GB,en;q=0.9',
            'Referer': 'https://www.basingstoke.gov.uk/',
            'Sec-Fetch-Dest': 'document',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-Site': 'cross-site',
            'Sec-Fetch-User': '?1',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.5845.188 Safari/537.36',
        }

        response = requests.get('https://www.basingstoke.gov.uk/bincollections', cookies=cookies,
                                 headers=headers, verify=False)

        if response.status_code != 200 or response.text == '0|error|500||':
            raise SystemError("Error retrieving data! Please try again or raise an issue on GitHub!")



        # Make a BS4 object
        soup = BeautifulSoup(response.text, features="html.parser")
        soup.prettify()

        bins = []

        for collection_type, collection_class in COLLECTION_KINDS.items():
            for date in soup.select(f"div#{collection_class} li"):
                bins.append({
                    "type": collection_type,
                    "collectionDate": datetime.strptime(
                        # Friday, 21 July 2023
                        date.get_text(strip=True),
                        '%A, %d %B %Y'
                    ).strftime(date_format)
                })

        return {
            "bins": bins
        }



from bs4 import BeautifulSoup, element
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        data = {"bins": []}
        collections = []

        # Parse the page and find all the result boxes
        soup = BeautifulSoup(page.text, features="html.parser")
        soup.prettify()
        results = soup.find_all("li", {"class": "resultListItem"})

        # If the result box has a wanted string in, we can use it. Check the contents of each box and find the
        # desired text and dates
        for r in results:
            if 'Your next' in r.text:
                if type(r.contents[10]) is element.NavigableString:
                    bin_text = r.contents[10].text.split(' ')[2].title() + ' bin'
                    bin_date = datetime.strptime(remove_ordinal_indicator_from_date_string(r.contents[11].text.strip()),
                                                 "%A %d %B %Y")
                else:
                    bin_text = r.contents[11].text.split(' ')[2].title() + ' bin'
                    bin_date = datetime.strptime(remove_ordinal_indicator_from_date_string(r.contents[12].text.strip()),
                                                 "%A %d %B %Y")

                if bin_date.date() >= datetime.now().date():
                    collections.append((bin_text, bin_date))

                # Sort the text and date elements by date
                ordered_data = sorted(collections, key=lambda x: x[1])

        # Put the elements into the dictionary
        for item in ordered_data:
            dict_data = {
                "type": item[0],
                "collectionDate": item[1].strftime(date_format),
            }
            data["bins"].append(dict_data)

        return data



import time
from typing import Any

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import TimeoutException, NoSuchElementException
from selenium.webdriver.support.ui import Select, WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from dateutil.parser import parse

from uk_bin_collection.uk_bin_collection.common import create_webdriver, date_format
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


class CouncilClass(AbstractGetBinDataClass):
    def wait_for_element_conditions(self, driver, conditions, timeout: int = 5):
        try:
            WebDriverWait(driver, timeout).until(conditions)
        except TimeoutException:
            print("Timed out waiting for page to load")
            raise

    def parse_data(self, page: str, **kwargs) -> dict:
        web_driver = kwargs.get("web_driver")
        page = "https://sevenoaks-dc-host01.oncreate.app/w/webpage/waste-collection-day"

        # Assign user info
        user_postcode = kwargs.get("postcode")
        user_paon = kwargs.get("paon")

        # Create Selenium webdriver
        driver = create_webdriver(web_driver)
        driver.get(page)

        # Enter postcode
        postcode_css_selector = "#address_search_postcode"
        self.wait_for_element_conditions(
            driver,
            EC.presence_of_element_located((By.CSS_SELECTOR, postcode_css_selector)),
        )
        postcode_input_box = driver.find_element(By.CSS_SELECTOR, postcode_css_selector)
        postcode_input_box.send_keys(user_postcode)
        postcode_input_box.send_keys(Keys.ENTER)

        # Select the dropdown
        self.wait_for_element_conditions(
            driver, EC.presence_of_element_located((By.XPATH, "//select/option[2]"))
        )
        select_address_dropdown = Select(driver.find_element(By.XPATH, "//select"))

        if user_paon is not None:
            for option in select_address_dropdown.options:
                if user_paon in option.text:
                    select_address_dropdown.select_by_visible_text(option.text)
                    break
        else:
            # If we've not been supplied an address, pick the second entry
            select_address_dropdown.select_by_index(1)

        # Grab the response blocks
        response_xpath_selector = "//div[@data-class_name]//h4/../../../.."
        self.wait_for_element_conditions(
            driver, EC.presence_of_element_located((By.XPATH, response_xpath_selector))
        )
        elements = driver.find_elements(By.XPATH, response_xpath_selector)

        # Iterate through them
        data = {"bins": []}
        for element in elements:
            try:
                raw_bin_name = element.find_element(By.TAG_NAME, "h4").text
                raw_next_collection_date = element.find_elements(
                    By.XPATH, ".//div[input]"
                )[1].text

                parsed_bin_date = parse(
                    raw_next_collection_date, fuzzy_with_tokens=True
                )[0]

                dict_data = {
                    "type": raw_bin_name,
                    "collectionDate": parsed_bin_date.strftime(date_format),
                }

                data["bins"].append(dict_data)

            except (IndexError, NoSuchElementException):
                print("Error finding element for bin")

        return data



from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Make a BS4 object
        soup = BeautifulSoup(page.text, features="lxml-xml")
        soup.prettify()

        data = {"bins": []}
        collections = []

        # Match bin types from API to their actual type
        bin_types = {
            "RESIDUAL BIN": "Grey bin",
            "RES 180": "Grey bin",
            "RES 240 STD": "Grey bin",
            "RES 360 STD+": "Grey bin",
            "RES 360 STD": "Grey bin",
            "RES 660": "Grey bin",
            "RES 770": "Grey bin",
            "RES BAG": "Grey bin",
            "RES 140 SML": "Grey bin",
            "REC 180 SML": "Blue bin",
            "REC 240 STD": "Blue bin",
            "REC 360": "Blue bin",
            "REC 770": "Blue bin",
            "MIXED REC 55 BOX": "Blue box",
            "PAPER 44 BOX": "Blue box",
            "PAPER BAG": "Paper bag",
            "ORG 180": "Brown bin",
            "ORG 240 STD": "Brown bin",
            "PAID ORGANIC": "Brown bin",
            "RES 1100": "Grey trade container",
            "REC GL 770": "Blue trade container"
        }
        # If the API errors, throw the exception
        if soup.find("Error") is not None:
            raise ConnectionAbortedError(soup.find("Error").text.strip())

        # Parse the XML and add to a list of collections
        for item in soup.find_all("BinRound"):
            try:
                bin_type = bin_types.get(item.find_next("Bin").text.replace("EMPTY BINS", "").strip())
                bin_date = datetime.strptime(item.find_next("DateTime").text, "%d/%m/%Y %H:%M:%S")
                if bin_date >= datetime.now():
                    collections.append((bin_type, bin_date))
            except:
                raise SystemError("Error has been encountered parsing API. Please try again later and if the issue "
                                  "persists, open a GitHub ticket!")

        # Sort the collections list by date
        ordered_data = sorted(collections, key=lambda x: x[1])

        # Put the elements into the dictionary
        for item in ordered_data:
            dict_data = {
                "type": item[0],
                "collectionDate": item[1].strftime(date_format),
            }
            data["bins"].append(dict_data)

        return data



import requests
from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import \
    AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber


class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Get postcode and UPRN from kwargs
        user_postcode = kwargs.get("postcode")
        user_uprn = kwargs.get("uprn")
        check_postcode(user_postcode)
        check_uprn(user_uprn)

        # Build URL to parse
        council_url = f"https://swale.gov.uk/bins-littering-and-the-environment/bins/collection-days?postcode={user_postcode.replace(' ', '+')}&addresses={user_uprn}&address-submit="

        # Parse URL and read if connection successful
        requests.packages.urllib3.disable_warnings()
        response = requests.get(council_url, verify=False)
        if response.status_code == 200:
            soup = BeautifulSoup(response.text, features="html.parser")
            soup.prettify()
        else:
            raise ConnectionAbortedError("Could not parse council website.")

        data = {"bins": []}

        # Get the collection bullet points on the page and parse them
        form_area = soup.find("form", {"class": "integration bin-lookup"})
        collections = [
            item.text.strip().split(",") for item in form_area.find_all("li")
        ]
        for c in collections:
            bin_type = c[0].strip()
            # temp_date = c[2].strip() + " " + str(datetime.now().year)
            bin_date = datetime.strptime(
                c[2].strip() + " " + str(datetime.now().year), "%d %B %Y"
            ).strftime(date_format)
            dict_data = {"type": bin_type, "collectionDate": bin_date}
            data["bins"].append(dict_data)

        return data
